#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "ping.h"

int ident;			/* process id to identify our packets */

int init4(char *target)
{
	struct sockaddr_in *to;
	struct hostent *hp;
	struct protoent *proto;

	if (!(proto = getprotobyname("icmp"))) {
		fprintf(stderr, "ping: unknown protocol icmp.\n");
		exit(1);
	}

	to=malloc(sizeof(struct sockaddr_in));
	afinet4_prot.whereto=(struct sockaddr *) to;
	memset(to, 0, sizeof(*to));

	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(target);

	/* FIXME: always true 
	 * to test: put a badly written entry in /etc/hosts 
	 */
	if (to->sin_addr.s_addr != -1)
		strncpy(hostname, target, MAXHOSTNAMELEN - 1);
	else {
		hp = gethostbyname2(target, AF_INET);
		if (!hp || !(hp->h_addr_list) || *(hp->h_addr_list[0])==-1) {
			free(to);
			return(-1);
		}

		to->sin_family = hp->h_addrtype;
		memcpy((caddr_t)&to->sin_addr, hp->h_addr, hp->h_length);
		strncpy(hostname, hp->h_name, MAXHOSTNAMELEN - 1);
	}

	if ((afinet4_prot.s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
		perror("ping: socket");
		exit(1);
	}

	sprintf(start_msg, "PING %s (%s): %d data bytes", 
		hostname,
		inet_ntoa(to->sin_addr),
		datalen);

	ident = getpid() & 0xFFFF;
	return(0);
}

/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 */
void build_icmp4(void *header, int len)
{
	register int nleft = len;
	register u_short *w = (u_short *)header;
	register int sum = 0;
	u_short answer = 0;
	struct icmphdr *icp = (struct icmphdr *)header;

	CLR(icp->icmp_seq % mx_dup_ck);

	icp->icmp_type = ICMP_ECHO;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = ntransmitted++;
	icp->icmp_id = ident;			/* ID */

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */

	icp->checksum = answer;
}


/*
 * pr_addr --
 *	Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
char *pr_addr4(struct sockaddr *add, int choice)
{
	struct sockaddr_in *addr= (struct sockaddr_in *) add;
	struct hostent *hp;
	static char buf[80];

	if ((options & F_NUMERIC) || (!choice) ||
	    !(hp = gethostbyaddr((char *)&addr->sin_addr, 4, AF_INET)))
		sprintf(buf, "%s", inet_ntoa(addr->sin_addr));
	else
		sprintf(buf, "%s (%s)", hp->h_name,
			inet_ntoa(addr->sin_addr));
	return(buf);
}


/*
 * pr_iph --
 *	Print an IP header with options.
 */
void pr_iph(struct iphdr *ip)
{
	int hlen;
	u_char *cp;

	hlen = ip->ip_hl << 2;
	cp = (u_char *)ip + 20;		/* point to options */

	printf("Vr HL TOS  Len   ID Flg  off TTL Pro  cks      Src      Dst Data\n");
	printf(" %1x  %1x  %02x %04x %04x",
	       ip->ip_v, ip->ip_hl, ip->ip_tos, ip->ip_len, ip->ip_id);
	printf("   %1x %04x", ((ip->ip_off) & 0xe000) >> 13,
	       (ip->ip_off) & 0x1fff);
	printf("  %02x  %02x %04x", ip->ip_ttl, ip->ip_p, ip->ip_sum);
	printf(" %s ", inet_ntoa(*(struct in_addr *)&ip->ip_src));
	printf(" %s ", inet_ntoa(*(struct in_addr *)&ip->ip_dst));
	
	/* dump and option bytes */
	while (hlen-- > 20) {
		printf("%02x", *cp++);
	}
	putchar('\n');
}


/*
 * pr_retip --
 *	Dump some info on a returned (via ICMP) IP packet.
 */
void pr_retip(struct iphdr *ip)
{
	int hlen;
	u_char *cp;

	pr_iph(ip);
	hlen = ip->ip_hl << 2;
	cp = (u_char *)ip + hlen;

	if (ip->ip_p == 6)
		printf("TCP: from port %u, to port %u (decimal)\n",
		       (*cp * 256 + *(cp + 1)), (*(cp + 2) * 256 + *(cp + 3)));
	else if (ip->ip_p == 17)
		printf("UDP: from port %u, to port %u (decimal)\n",
		       (*cp * 256 + *(cp + 1)), (*(cp + 2) * 256 + *(cp + 3)));
}



/*
 * pr_icmph --
 *	Print a descriptive string about an ICMP header.
 */
void pr_icmp4h(void *icmp_header)
{
	struct icmphdr *icp=(struct icmphdr *)icmp_header;

	switch(icp->icmp_type) {
	case ICMP_ECHOREPLY:
		printf("Echo Reply\n");
		/* XXX ID + Seq + Data */
		break;
	case ICMP_DEST_UNREACH:
		switch(icp->icmp_code) {
		case ICMP_NET_UNREACH:
			printf("Destination Net Unreachable\n");
			break;
		case ICMP_HOST_UNREACH:
			printf("Destination Host Unreachable\n");
			break;
		case ICMP_PROT_UNREACH:
			printf("Destination Protocol Unreachable\n");
			break;
		case ICMP_PORT_UNREACH:
			printf("Destination Port Unreachable\n");
			break;
		case ICMP_FRAG_NEEDED:
			printf("frag needed and DF set\n");
			break;
		case ICMP_SR_FAILED:
			printf("Source Route Failed\n");
			break;
		default:
			printf("Dest Unreachable, Bad Code: %d\n",
			       icp->icmp_code);
			break;
		}
		
		/* Print returned IP header information */
#ifndef icmp_data
		pr_retip(icp + 1);
#else
		pr_retip((struct ip *)icp->icmp_data);
#endif
		break;
	case ICMP_SOURCE_QUENCH:
		printf("Source Quench\n");
#ifndef icmp_data
		pr_retip(icp + 1);
#else
		pr_retip((struct ip *)icp->icmp_data);
#endif
		break;
	case ICMP_REDIRECT:
		switch(icp->icmp_code) {
		case ICMP_REDIR_NET:
			printf("Redirect Network");
			break;
		case ICMP_REDIR_HOST:
			printf("Redirect Host");
			break;
		case ICMP_REDIR_NETTOS:
			printf("Redirect Type of Service and Network");
			break;
		case ICMP_REDIR_HOSTTOS:
			printf("Redirect Type of Service and Host");
			break;
		default:
			printf("Redirect, Bad Code: %d", icp->icmp_code);
			break;
		}
		printf("(New addr: 0x%08lx)\n", (long)icp->icmp_gwaddr);
#ifndef icmp_data
		pr_retip(icp + 1);
#else
		pr_retip((struct ip *)icp->icmp_data);
#endif
		break;
	case ICMP_ECHO:
		printf("Echo Request\n");
		/* XXX ID + Seq + Data */
		break;
	case ICMP_TIME_EXCEEDED:
		switch(icp->icmp_code) {
		case ICMP_EXC_TTL:
			printf("Time to live exceeded\n");
			break;
		case ICMP_EXC_FRAGTIME:
			printf("Frag reassembly time exceeded\n");
			break;
		default:
			printf("Time exceeded, Bad Code: %d\n",
			       icp->icmp_code);
			break;
		}
#ifndef icmp_data
		pr_retip(icp + 1);
#else
		pr_retip((struct ip *)icp->icmp_data);
#endif
		break;
	case ICMP_PARAMETERPROB:
		printf("Parameter problem: pointer = 0x%02x\n",
		       icp->icmp_gwaddr);
#ifndef icmp_data
		pr_retip(icp + 1);
#else
		pr_retip((struct ip *)icp->icmp_data);
#endif
		break;
	case ICMP_TIMESTAMP:
		printf("Timestamp\n");
		/* XXX ID + Seq + 3 timestamps */
		break;
	case ICMP_TIMESTAMPREPLY:
		printf("Timestamp Reply\n");
		/* XXX ID + Seq + 3 timestamps */
		break;
	case ICMP_INFO_REQUEST:
		printf("Information Request\n");
		/* XXX ID + Seq */
		break;
	case ICMP_INFO_REPLY:
		printf("Information Reply\n");
		/* XXX ID + Seq */
		break;
#ifdef ICMP_MASKREQ
	case ICMP_MASKREQ:
		printf("Address Mask Request\n");
		break;
#endif
#ifdef ICMP_MASKREPLY
	case ICMP_MASKREPLY:
		printf("Address Mask Reply\n");
		break;
#endif
	default:
		printf("Bad ICMP type: %d\n", icp->icmp_type);
	}
}

/*
 * Returns the ICMP header, else NULL
 */
void* get_icmp4_hdr(char *packet, int *len, int *ttl)
{
	*len = ((struct iphdr *)packet)->ip_hl << 2;	
	*ttl = ((struct iphdr *)packet)->ip_ttl;	

	return((void *)(packet + *len));
}


/*
 * Analyse IPv4 specific options
 */
void read_options4()
{
#ifdef IP_OPTIONS
	char rspace[3 + 4 * NROUTES + 1];	/* record route space */
#endif
	
	/* record route option */
	if (options & F_RROUTE) {
#ifdef IP_OPTIONS
		rspace[IPOPT_OPTVAL] = IPOPT_RR;
		rspace[IPOPT_OLEN] = sizeof(rspace)-1;
		rspace[IPOPT_OFFSET] = IPOPT_MINOFF;
		if (setsockopt(afinet4_prot.s, IPPROTO_IP, IP_OPTIONS, rspace,
			       sizeof(rspace)) < 0) {
			perror("ping: record route");
			exit(1);
		}
#else
		fprintf(stderr, "ping: record route not available in this implementation.\n");
		exit(1);
#endif /* IP_OPTIONS */
	}
		
#ifdef 0
	if (options & MULTICAST_NOLOOP) {
		if (setsockopt(afinet4_prot->s, IPPROTO_IP, IP_MULTICAST_LOOP,
			       &loop, 1) == -1) {
			perror ("can't disable multicast loopback");
			exit(92);
		}
	}
	if (options & MULTICAST_TTL) {
		if (setsockopt(afinet4_prot->s, IPPROTO_IP, IP_MULTICAST_TTL,
			       &ttl, 1) == -1) {
			perror ("can't set multicast time-to-live");
			exit(93);
		}
	}
	if (options & MULTICAST_IF) {
		if (setsockopt(afinet4_prot->s, IPPROTO_IP, IP_MULTICAST_IF,
			       &ifaddr, sizeof(ifaddr)) == -1) {
			perror ("can't set multicast source interface");
			exit(94);
		}
	}
#endif

}

struct proto_ops afinet4_prot = {
	AF_INET,
	0,
	sizeof(struct sockaddr_in),
        NULL,
	init4,
	build_icmp4,
	pr_pack,
	read_options4,
        get_icmp4_hdr,
	pr_icmp4h,
	pr_addr4};


