#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <linux/ipv6.h>
#include <linux/icmpv6.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "ping.h"

int ident;			/* process id to identify our packets */

int init6(char *target)
{
unsigned int	i;

	struct sockaddr_in6 *to;
	struct protoent *proto;

	struct hostent *host;
	int offset;
	char buff[INET6_ADDRSTRLEN];

	if (!(proto = getprotobyname("icmp6"))) {
		fprintf(stderr, "ping: unknown protocol icmp6 (see /etc/protocols)\n");
		exit(1);
	}
	to=malloc (sizeof(struct sockaddr_in6));
	afinet6_prot.whereto=(struct sockaddr *) to;

	memset(to, 0, sizeof(struct sockaddr_in6));

	to->sin6_family = AF_INET6;
	to->sin6_port = IPPROTO_ICMPV6;

	if (inet_pton(AF_INET6, target, &to->sin6_addr) == 1)
		strncpy(hostname, target, MAXHOSTNAMELEN - 1);
	else {
		if (strstr (target, "::")) {
			target += 2; /* skip the semicolon */
			options |= F_IP4COMP;
		}

        	if ((host = gethostbyname2 (target, AF_INET6))) {
	   	    for (i = 0; host->h_addr_list[i]; i++) {
			if (((options & F_IP4COMP) && IS_IPV4ADDR6( *(struct in6_addr*) host->h_addr_list[i])) ||
		   	   (!(options & F_IP4COMP) && !IS_IPV4ADDR6(*(struct in6_addr *) host->h_addr_list[i])))  { 
		   	   COPY_ADDR6(to->sin6_addr, *(struct in6_addr *) host->h_addr_list[i]);
		   	   break;
		        } 
	            }
		    if (!host->h_addr_list[i]) {/* no ipv6 address found */
			perror("ping: no address found");
			exit (1);
		     }
	        } else {
	                if (!((options & F_IP4COMP) && (host = gethostbyname2 (target, AF_INET)))) {
			     free(to);
			     return(-1);
			}
    	               CLR_ADDR6(to->sin6_addr); 
	    	       to->sin6_addr.s6_addr32[3] = *(__u32 *) host->h_addr_list[0];
		}
                strncpy(hostname, host->h_name, MAXHOSTNAMELEN - 1);
	}
	
	if ((afinet6_prot.s = socket(AF_INET6, SOCK_RAW, proto->p_proto)) < 0) {
	  perror("ping: socket not created");
	  exit(1);
	}

	/* Checksum must be at offset 2 in icmp header */
	offset=2;            
	if (setsockopt(afinet6_prot.s, IPPROTO_IPV6, IPV6_CHECKSUM, 
		       &offset, sizeof(offset)) < 0) {
		perror("ping: unable to set Raw checksum");
		exit(1);
	}

	inet_ntop(AF_INET6, &to->sin6_addr,
		  buff, INET6_ADDRSTRLEN);

	sprintf(start_msg, "PING %s (%s): %d data bytes\n", 
		hostname,
		buff,
		datalen);

	ident = getpid() & 0xFFFF;
	return(0);
}

/*
 * Returns the ICMP header, else NULL.
 * len is the length of the icmp header
 */
void* get_icmp6_hdr(char *packet, int *len, int *ttl)
{
	struct ipv6hdr *hdr = (struct ipv6hdr *)packet;
	struct icmpv6 *icmphdr;

	*len = sizeof(struct ipv6hdr);
	*ttl = hdr->hop_limit;
	icmphdr=(struct icmpv6 *)(hdr+1);
	
	/* FIXME: if jumbo datagram, the size is 0, and the function exits. */
	
	while (hdr->next_hdr != IPPROTO_ICMPV6) {
		switch (hdr->next_hdr) {
			
		default:
			fprintf(stderr, "ping: unknown next header %2x\n", hdr->next_hdr);
			exit(1);
		}
	}
	
	if (*len>0)
		return(icmphdr);
	else
		return(NULL);
}

/*
 * Analyse IPv6 specific options
 */
void read_options6()
{
#ifdef 0
  if (options & MULTICAST_NOLOOP) {
                if (setsockopt(afinet6_prot.s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
                               &loop, 1) == -1) {
                        perror ("can't disable multicast loopback");
                        exit(92);
                }
  }
#endif

  if (options & MULTICAST_TTL) {
                if (setsockopt(afinet6_prot.s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
                               &ttl, 1) == -1) {
                        perror ("can't set multicast hop limit");
                        exit(93);
                }
                if (setsockopt(afinet6_prot.s, IPPROTO_IPV6, IPV6_UNICAST_HOPS,
                               &ttl, 1) == -1) {
                        perror ("can't set unicast hop limit");
                        exit(93);
                }

  }
#ifdef 0 /* FIXME Pass the address.... */
  if (options & MULTICAST_IF) {
                if (setsockopt(afinet6_prot->s, IPPROTO_IPV6,IPV6_MULTICAST_IF,
                               &ifaddr, sizeof(ifaddr)) == -1) {
                        perror ("can't set multicast source interface");
                        exit(94);
                }
  }
#endif

}


/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 */
void build_icmp6(void *header, int len)
{
	struct icmpv6 *icp = (struct icmpv6 *)header;

	CLR(icp->icmp6_seq % mx_dup_ck);

	icp->icmp6_type = ICMP6_ECHO;
	icp->icmp6_code = 0;
	icp->icmp6_seq = ntransmitted++;
	icp->icmp6_id = ident;			/* ID */
}


/*
 * Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
char *pr_addr6(struct sockaddr *from, int choice)
{
	struct sockaddr_in6 *addr= (struct sockaddr_in6 *) from;
	struct hostent *hp;
	static char buf[200], buf2[200];
	
	if ( (options & F_NUMERIC) || (!choice) ||
	    !(hp = gethostbyaddr((char *)&addr->sin6_addr, 
				 sizeof(struct in6_addr), AF_INET6)))
	{
		inet_ntop(AF_INET6, &addr->sin6_addr, 
			  buf, sizeof(buf));
	}
	else
	{
		sprintf(buf, "%s (%s)", hp->h_name, 
			inet_ntop(AF_INET6, &addr->sin6_addr, 
				  buf2, sizeof(struct in6_addr)));
	}
	return(buf);
}


struct proto_ops afinet6_prot = {
        AF_INET6,
	0,
	sizeof(struct sockaddr_in6),
        NULL,
	init6,
	build_icmp6,
	pr_pack,
	read_options6,
        get_icmp6_hdr,
        NULL,
        pr_addr6};
















