/*
 * PING   :     Modifications for IPv6 
 *
 * Authors:     Frank Zago <zago@masi.ibp.fr>
 *              INRIA      <Christophe.Diot@sophia.inria.fr>
 *              MASI       <Eric.Horlait@masi.ibp.fr>
 *              This software has been developped with the financial support
 *              of DRET (French Military Research Agency).
 *
 *              $Id: ping.h,v 1.2 1996/11/05 17:33:58 eh Exp $ 
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *              This notice must be copied with the distributed package.
 */

/* 
 * That structure defines how ping must act with a given protocol, 
 * currently PF_INET and PF_INET6.
 */
   
struct proto_ops {
	int prot_af;			/* adress family */
	int s;				/* socket descriptor */
	int addrlen;			/* address length */
	struct sockaddr *whereto;	/* who to ping */
	int (*init)(char *target);	/* init protocol */
	void (*build_cm_packet)(void *header, /* build an icmp packet */
				int len);
	void (*pr_pack)(char *buf,	/* print an incoming packet */
			int cc,
			struct sockaddr *from);
	void (*options)(void);
	void *(*get_icmp_hdr)(char *packet, int *len, int *ttl);
	void (*pr_icmp_hdr)(void *);            /* print an icmp header */
	char *(*pr_addr)(struct sockaddr *, int choice);    /* ascii host adresss */
#ifdef 0
	void (*pr_retip)(void *);		/* dump info on IP packet */
#endif
};

void pr_pack(char *buf, int cc, struct sockaddr *frm);

/* 
 * Various options
 */
#define ICMP_MINLEN	28

#define	DEFDATALEN	(64 - 8)	/* default data length */
#define	MAXIPLEN	60
#define	MAXICMPLEN	76
#define	MAXPACKET	(65536 - 60 - 8)/* max packet size */
#define	MAXWAIT		10		/* max seconds to wait for response */
#define	NROUTES		9		/* number of record route slots */

#define	A(bit)		rcvd_tbl[(bit)>>3]	/* identify byte in array */
#define	B(bit)		(1 << ((bit) & 0x07))	/* identify bit in byte */
#define	SET(bit)	(A(bit) |= B(bit))
#define	CLR(bit)	(A(bit) &= (~B(bit)))
#define	TST(bit)	(A(bit) & B(bit))

/*
 * MAX_DUP_CHK is the number of bits in received table, i.e. the maximum
 * number of received sequence numbers we can keep track of.  Change 128
 * to 8192 for complete accuracy...
 */
#define	MAX_DUP_CHK	(8 * 128)

/* 
 * Command line options 
 */
#define	F_FLOOD			0x0001
#define	F_INTERVAL		0x0002
#define	F_NUMERIC		0x0004
#define	F_PINGFILLED		0x0008
#define	F_QUIET			0x0010
#define	F_RROUTE		0x0020
#define	F_SO_DEBUG		0x0040
#define	F_SO_DONTROUTE		0x0080
#define	F_VERBOSE		0x0100
#define F_IP4COMP		0x0200

/* multicast options */
#define MULTICAST_NOLOOP	0x1000
#define MULTICAST_TTL		0x2000
#define MULTICAST_IF		0x4000

/*
 * Some shorcuts
 */
#define icmp_type type
#define icmp_code code
#define icmp_cksum checksum
#define icmp_id un.echo.id
#define icmp_seq un.echo.sequence
#define icmp_gwaddr un.gateway
#define ip_hl ihl
#define ip_v version
#define ip_tos tos
#define ip_len tot_len
#define ip_id id
#define ip_off frag_off
#define ip_ttl ttl
#define ip_p protocol
#define ip_sum check
#define ip_src saddr
#define ip_dst daddr

void pinger(void);
void usage(void);
void fill(char *bp, char *patp);
void tvsub(struct timeval *out, struct timeval *in);

extern struct proto_ops *prot;
extern struct proto_ops afinet4_prot;
extern struct proto_ops afinet6_prot;

extern char hostname[];

extern long ntransmitted;	
extern int ident;	
extern char start_msg[];
extern int datalen;
extern int options;

extern long nreceived;	  
extern long nrepeats;

extern int timing;
extern long tmin;
extern long tmax;
extern u_long tsum;

int mx_dup_ck;
extern char rcvd_tbl[];

extern char BSPACE;

extern u_char outpack[];
extern u_char ttl;
