/*-
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Van Jacobson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1990, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)traceroute.c	8.1 (Berkeley) 6/6/93";
#endif /* not lint */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#if __linux__
#include <endian.h>
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>

#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "traceroute.h"

/*
 * Format of user data 
 */
 struct payload {
 	struct timeval tv;	/* time packet  */
	u_char seq;		/* sequence number of this packet left*/
	u_char ttl;		/* ttl packet left with */
 };
/*
 * format of a (udp) probe packet.
 */
struct opacket {
	struct ip ip;
	struct udphdr udp;
	struct payload data;
};

int init4 __P((char *));
void send_probe4 __P((struct proto_ops *, u_char *, int, int));
int recv_packet4 __P((struct proto_ops *, u_char *, int, int *, struct timeval *));
void getnewhop4 __P((struct proto_ops *, int, double, int));
char *inetname __P((struct in_addr));
char *pr_type4 __P((u_char));


static char printable_localaddr4[INET_ADDRSTRLEN];

int
init4(char * target)
{
	struct sockaddr_in *from, *to;
	struct in_addr *lastaddr;
	struct hostent *hp;
	struct protoent *pe;
	char printable_address[INET_ADDRSTRLEN];
	int i;

	if (!(to =  malloc(sizeof(struct sockaddr_in)))){
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	afinet4_prot.whereto = (struct sockaddr *) to;
	memset ((char *)to, 0, sizeof(struct sockaddr_in));

	if ((pe = getprotobyname("icmp")) == NULL) {
		Fprintf(stderr, "icmp: unknown protocol\n");
		exit(10);
	}

	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(target);
	if (to->sin_addr.s_addr != -1)
		strncpy (hostname, target, MAXHOSTNAMELEN -1);
	else {
		hp = gethostbyname2(target, AF_INET);
		if (!hp || !(hp->h_addr_list) || *(hp->h_addr_list[0]) == -1) {
			free (to);
			return(-1);
		} else {
			to->sin_family = hp->h_addrtype;
			bcopy(hp->h_addr, (caddr_t)&to->sin_addr, hp->h_length);
			strncpy (hostname, hp->h_name, MAXHOSTNAMELEN -1);
		}
	}

	if ((afinet4_prot.s = socket(AF_INET, SOCK_RAW, pe->p_proto)) < 0) {
		perror("traceroute: icmp socket");
		exit(5);
	}

	if ((afinet4_prot.sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
		perror("traceroute: raw socket");
		exit(5);
	}
#ifdef IP_HDRINCL
	/* send with the ip header */
	if (setsockopt(afinet4_prot.sndsock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
		       sizeof(on)) < 0) {
		perror("traceroute: IP_HDRINCL");
		exit(6);
	}
#endif IP_HDRINCL

	if (!(lastaddr =  malloc(sizeof(struct in_addr)))){
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	memset ((char *)lastaddr, 0, sizeof (struct in_addr));
	afinet4_prot.lastaddr = (void *) lastaddr;
	
	if (!(from =  malloc(sizeof(struct sockaddr_in)))){
			fprintf(stderr, "malloc failed\n");
			exit(1);
	}
	afinet4_prot.wherefrom = (struct sockaddr *) from;
	memset ((char *)from, 0, sizeof(struct sockaddr_in));

	if (source) {
		from->sin_family = AF_INET;
		if (inet_pton(AF_INET, source, &from->sin_addr)){
			Printf("traceroute: unknown addr %s\n", source);
			exit(1);
		}
#ifndef IP_HDRINCL
		if (bind(afinet4_prot.sndsock, (struct sockaddr *)from, sizeof(*from)) < 0){
			perror ("traceroute: bind");
			exit (1);
		}
#endif IP_HDRINCL
	} 
	Fprintf(stderr, "traceroute to %s (%s)", hostname,
			inet_ntop(AF_INET, &to->sin_addr, printable_address, INET_ADDRSTRLEN));
	return (0);

}

void
fillip4 (struct proto_ops *prot, u_char *outpacket)
{
	struct opacket *op;
	struct sockaddr_in  *to, *from;
	
	(void) memset (outpacket, 0, datalen);
	op = (struct opacket *) outpacket;
	to = (struct sockaddr_in *)prot->whereto;
	from = (struct sockaddr_in *)prot->wherefrom;

	op->ip.ip_dst = to->sin_addr;
	op->ip.ip_tos = tos;
	op->ip.ip_v = IPVERSION;
	op->ip.ip_id = 0;
	op->ip.ip_src = from->sin_addr;
	op->ip.ip_off = 0;
	op->ip.ip_p = IPPROTO_UDP;
}


void
send_probe4(struct proto_ops *prot, u_char *outpacket, int seq, int ttl)
{
	struct opacket *op = (struct opacket *) outpacket;
	struct ip *ip = &op->ip;
	struct udphdr *up = &op->udp;
	int nbcharsend;
	struct timezone tz;

	ip->ip_hl = sizeof(*ip) >> 2; /* header length in word of 32 bits */
	ip->ip_len = datalen;
	ip->ip_ttl = ttl;
	ip->ip_id = htons(ident+seq);

	up->uh_sport = htons(ident);
	up->uh_dport = htons(port+seq);
	up->uh_ulen = htons((u_short)(datalen - sizeof(struct ip)));
	up->uh_sum = 0;

	op->data.seq = seq;
	op->data.ttl = ttl;
	(void) gettimeofday(&op->data.tv, &tz);

	nbcharsend = sendto(prot->sndsock, (char *)outpacket, datalen, 0, prot->whereto, sizeof(struct sockaddr_in));
	if (nbcharsend < 0 || nbcharsend != datalen)  {
		if (nbcharsend < 0)
			perror("sendto");
		Printf("traceroute: wrote %s %d chars, ret=%d\n", hostname, datalen, nbcharsend);
		(void) fflush(stdout);
	}
}




int
recv_packet4(struct proto_ops *prot, u_char *packet, int seq, int *cc, struct timeval *tv)
{
	struct icmp *icp;
	u_char type, code;
	int hlen;
	
#ifndef ARCHAIC
	struct ip *ip;
	struct sockaddr_in *from;

	from = (struct sockaddr_in *) prot->wherefrom;
	ip = (struct ip *) packet;
	if (!inet_ntop(AF_INET, &ip->ip_dst, printable_localaddr4, INET_ADDRSTRLEN)) {
		perror ("inet_ntop of destination address: ");
		printable_localaddr4 [0] = 0;
	}

	hlen = ip->ip_hl << 2;
	if (*cc < hlen + ICMP_MINLEN) {
		if (verbose)
			Printf("packet too short (%d bytes) from %s\n", *cc,
				inet_ntoa(from->sin_addr));
		return (0);
	}
	*cc -= hlen;
	icp = (struct icmp *)(packet + hlen);
#else
	icp = (struct icmp *)packet;
#endif ARCHAIC

	type = icp->icmp_type; 
	code = icp->icmp_code;
	if ((type == ICMP_TIMXCEED && code == ICMP_TIMXCEED_INTRANS) ||
	    type == ICMP_UNREACH) {
		struct ip *hip;
		struct udphdr *up;

		hip = &icp->icmp_ip;
		hlen = hip->ip_hl << 2;
		up = (struct udphdr *)((u_char *)hip + hlen);

		if (hlen + 12 <= *cc && hip->ip_p == IPPROTO_UDP &&
		    up->uh_sport == htons(ident) &&
		    up->uh_dport == htons(port+seq)) {
			/* struct payload *pp;
			
			pp = (struct payload *) (up +1);
			*tv = pp->tv; */
			return (type == ICMP_TIMXCEED? -1 : code+1);
		}
	}
#ifndef ARCHAIC
	if (verbose) {
		int i;
		u_long *lp = (u_long *)&icp->icmp_ip;

		Printf("\n%d bytes from %s to %s", *cc,
			inet_ntoa(from->sin_addr), inet_ntoa(ip->ip_dst));
		Printf(": icmp type %d (%s) code %d\n", type, pr_type4(type),
		       icp->icmp_code);

		/* Dump IP Packet contained in the ICMP packet received */
		for (i = 4; i < *cc ; i += sizeof(long))
			Printf("%2d: x%8.8lx\n", i, *lp++);
	}

#endif ARCHAIC
	return(0);
}


void 
getnewhop4 (struct proto_ops *prot, int code, double difftime, int cc)
{
	struct sockaddr_in *from;
	struct in_addr *lastaddr;
	struct ip *ip;

	from = (struct sockaddr_in *) prot->wherefrom;
	lastaddr = (struct in_addr *) prot->lastaddr;
	
	if (from->sin_addr.s_addr != lastaddr->s_addr) {
		memcpy (prot->lastaddr, &from->sin_addr, sizeof (struct in_addr));

		if (nflag)
			Printf(" %s", inet_ntoa(from->sin_addr));
		else
			Printf(" %s (%s)", inetname(from->sin_addr),
		       			inet_ntoa(from->sin_addr));
		if (verbose)
			printf (" %d bytes to %s", cc, printable_localaddr4);
	}
	Printf("  %g ms", difftime);
	switch(code - 1) {
			case ICMP_UNREACH_PORT:
#ifndef ARCHAIC
				ip = (struct ip *)inpacket;
				if (ip->ip_ttl <= 1)
					Printf(" !");
#endif ARCHAIC
				++got_there;
				break;
			case ICMP_UNREACH_NET:
				++unreachable;
				Printf(" !N");
				break;
			case ICMP_UNREACH_HOST:
				++unreachable;
				Printf(" !H");
				break;
			case ICMP_UNREACH_PROTOCOL:
				++got_there;
				Printf(" !P");
				break;
			case ICMP_UNREACH_NEEDFRAG:
				++unreachable;
				Printf(" !F");
				break;
			case ICMP_UNREACH_SRCFAIL:
				++unreachable;
				Printf(" !S");
				break;
		}
}



/*
 * Convert an ICMP "type" field to a printable string.
 */
char *
pr_type4(t)
	u_char t;
{
	static char *ttab[] = {
	"Echo Reply",	"ICMP 1",	"ICMP 2",	"Dest Unreachable",
	"Source Quench", "Redirect",	"ICMP 6",	"ICMP 7",
	"Echo",		"ICMP 9",	"ICMP 10",	"Time Exceeded",
	"Param Problem", "Timestamp",	"Timestamp Reply", "Info Request",
	"Info Reply"
	};

	if(t > 16)
		return("OUT-OF-RANGE");

	return(ttab[t]);
}


/*
 * Construct an Internet address representation.
 * If the nflag has been supplied, give
 * numeric value, otherwise try for symbolic name.
 */
char *
inetname(struct in_addr in)
{
	register char *cp;
	static char line[50];
	struct hostent *hp;
	static char domain[MAXHOSTNAMELEN + 1];
	static int first = 1;

	if (first && !nflag) {
		first = 0;
		if (gethostname(domain, MAXHOSTNAMELEN) == 0 &&
		    (cp = index(domain, '.')))
			(void) strcpy(domain, cp + 1);
		else
			domain[0] = 0;
	}
	cp = 0;
	if (!nflag && in.s_addr != INADDR_ANY) {
		hp = gethostbyaddr((char *)&in, sizeof (in), AF_INET);
		if (hp) {
			if ((cp = index(hp->h_name, '.')) &&
			    !strcmp(cp + 1, domain))
				*cp = 0;
			cp = (char *)hp->h_name;
		}
	}
	if (cp)
		(void) strcpy(line, cp);
	else {
		in.s_addr = ntohl(in.s_addr);
#define C(x)	((x) & 0xff)
		Sprintf(line, "%lu.%lu.%lu.%lu", C(in.s_addr >> 24),
			C(in.s_addr >> 16), C(in.s_addr >> 8), C(in.s_addr));
	}
	return (line);
}



#ifdef notyet
/*
 * Checksum routine for Internet Protocol family headers (C Version)
 */
u_short
in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1)
		sum += *(u_char *)w;

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return (answer);
}
#endif notyet

struct proto_ops afinet4_prot = {
	AF_INET,
	0,
	0,
	sizeof(struct sockaddr_in),
	sizeof(struct opacket),
	NULL,
	NULL,
	NULL,
	init4,
	send_probe4,
	recv_packet4,
	getnewhop4,
	fillip4};
