/*-
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Van Jacobson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1990, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)traceroute.c	8.1 (Berkeley) 6/6/93";
#endif /* not lint */

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#if __linux__
#include <endian.h>
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>

#include <linux/in6.h>
#include <linux/ipv6.h>
#include <linux/icmpv6.h>

#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "traceroute.h"


struct pkt_format {
	unsigned long ident;
	unsigned long seq;
	struct timeval tv;
};


char * pr_type __P((u_char t));
int    init6  __P((char *));
void   send_probe6  __P((struct proto_ops *, u_char *, int , int ));
struct icmpv6_error* get_icmp_hdr __P((char *));
int    recv_packet6  __P((struct proto_ops *, u_char *, int , int *, struct timeval *));
void   getnewhop6  __P((struct proto_ops *, int , double , int ));

static char    printable_localaddr6[INET6_ADDRSTRLEN];

static void dump_hex(char *buf, int nb)
{
   int i = 0, n = 0;

   for (; nb > 0; nb--) {
      printf("%02hx ", (unsigned char) buf[i]);
      i++;
      n++;
      if (n == 16) {
	 printf("\n");
	 n = 0;
      }
   }
   if (n != 0) {
      printf("\n");
   }
}		


int 
init6(char *target)
{
	struct sockaddr_in6 *to, *from ;
	struct in6_addr *lastaddr6;
	struct hostent	*host;
	struct protoent *proto;
	char printable_address[INET6_ADDRSTRLEN];
	int i;

	if (!(to =  malloc(sizeof(struct sockaddr_in6)))){
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	afinet6_prot.whereto = (struct sockaddr *) to;
	memset ((char *)to, 0, sizeof(struct sockaddr_in6));

	if ((proto = getprotobyname("icmp6")) == NULL) {
		Fprintf(stderr, "icmp6: unknown protocol\n");
		exit(10);
	}

	to->sin6_family = AF_INET6;
	to->sin6_port = htons(port);

	if (inet_pton(AF_INET6, target, &to->sin6_addr))
		/* target is a valide ipv6 address */
		strncpy (hostname, target, MAXHOSTNAMELEN -1);
	else 
	{
		if (strstr (target, "::")) {
			target += 2; /* skip the semicolon */
			options |= F_IP4COMP;
		}

        	if ((host = gethostbyname2 (target, AF_INET6))) {
	   	    for (i = 0; host->h_addr_list[i]; i++) {
			if (((options & F_IP4COMP) && IS_IPV4ADDR6( *(struct in6_addr*) host->h_addr_list[i])) ||
		   	   (!(options & F_IP4COMP) && !IS_IPV4ADDR6(*(struct in6_addr*) host->h_addr_list[i])))  { 
		   	   COPY_ADDR6(to->sin6_addr, *(struct in6_addr *) host->h_addr_list[i]);
		   	   break;
		        } 
	            }
		    if (!host->h_addr_list[i]) {/* no ipv6 address found */
			perror("traceroute: no address found");
			exit (1);
		     }
	        } else {
	                if (!((options & F_IP4COMP) && (host = gethostbyname2 (target, AF_INET)))) {
			     free(to);
			     return(-1);
			}
    	               CLR_ADDR6(to->sin6_addr); 
	    	       to->sin6_addr.s6_addr32[3] = *(__u32 *) host->h_addr_list[0];
		}
	        strncpy(hostname, host->h_name, MAXHOSTNAMELEN - 1);
	}


	if ((afinet6_prot.s = socket(AF_INET6, SOCK_RAW, proto->p_proto)) < 0) {
		perror("traceroute: icmpv6 socket");
		exit(5);
	}

	if ((afinet6_prot.sndsock = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("traceroute: UDP socket");
		exit(5);
	}

	if (!(lastaddr6 =  malloc(sizeof(struct in6_addr)))){
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	CLR_ADDR6 (*lastaddr6)
	afinet6_prot.lastaddr = (void *) lastaddr6;
	
	if (!(from =  malloc(sizeof(struct sockaddr_in6)))){
			fprintf(stderr, "malloc failed\n");
			exit(1);
	}
	afinet6_prot.wherefrom = (struct sockaddr *) from;
	memset ((char *)from, 0, sizeof(struct sockaddr_in6));
	
	if (source) {
		from->sin6_family = AF_INET6;
		if (!(inet_pton(AF_INET6, source, &from->sin6_addr))){
			Printf("traceroute: unknown addr %s\n", source);
			exit(1);
		}
		if (bind(afinet6_prot.sndsock, (struct sockaddr *)from, sizeof(*from)) < 0){
			perror ("traceroute: bind");
			exit (1);
		}
	}
	
	Fprintf(stderr, "traceroute to %s (%s)", hostname,
		inet_ntop(AF_INET6, &to->sin6_addr, printable_address, INET6_ADDRSTRLEN));
	return (0);
}


void 
send_probe6(struct proto_ops *prot, u_char *outpacket, int seq, int ttl)
{
	struct pkt_format *pkt = (struct pkt_format *) outpacket;
	int    nbcharsend;
	struct timezone tz;		/* leftover */
	struct sockaddr_in6 *to = (struct sockaddr_in6 *) prot->whereto;

	to->sin6_port = htons(port+seq); 
	pkt->ident = htonl(ident);
	pkt->seq = htonl(seq);
	gettimeofday(&pkt->tv, &tz);

	if (setsockopt(prot->sndsock, SOL_IPV6, IPV6_UNICAST_HOPS, &ttl, sizeof(int)) <0) {
		perror("setsockopt");
		exit(1);
	}
        /* dump_hex((char *)outpacket, datalen); */
	nbcharsend = sendto(prot->sndsock, outpacket, datalen, 0,
		   prot->whereto, sizeof(struct sockaddr_in6));

	if (nbcharsend < 0 || nbcharsend != datalen)  {
		if (nbcharsend<0)
			perror("sendto");
		Printf("traceroute: wrote %s %d chars, ret=%d\n", hostname,
			datalen, nbcharsend);
		(void) fflush(stdout);
	}
}



/*
 * Returns the ICMP header,
 */
struct icmpv6_error* get_icmp_hdr(char *packet)
{
	struct ipv6hdr *hdr = (struct ipv6hdr *)packet;
	struct icmpv6_error *icmphdr;
	int header_len, len;
	unsigned char next_header;
	char * new_header;

	len = hdr->payload;
	next_header = hdr->next_hdr;
	new_header = (char *)(hdr+1);
	header_len = 0;
	
	/* FIXME: if jumbo datagram, the size is 0, and the function exits. */
	
	do {
		switch (next_header) {
			case IPPROTO_ICMPV6:	 /* ICMPv6 */
				icmphdr=(struct icmpv6_error *)new_header;
				break;
			case IPPROTO_HOPOPTS: 	      /* IPv6 hop-by-hop options */
			case IPPROTO_ROUTING:         /* IPv6 routing header */
			case IPPROTO_FRAGMENT:        /* IPv6 fragmentation header */
			case IPPROTO_ESP:             /* encapsulating security payload */
			case IPPROTO_AH:              /* authentication header */
			case IPPROTO_NONE:            /* IPv6 no next header */
			case IPPROTO_DSTOPTS: 
			default:
				fprintf(stderr, "traceroute: unknown next header %2x\n",
						 next_header);
				exit(1);
		}
	} while (next_header != IPPROTO_ICMPV6);
	
	if (len>0)
		return(icmphdr);
		
	fprintf (stderr, "traceroute: Packet too short, icmp packet not valide ");
	exit (1);
}

int 
recv_packet6(struct proto_ops *prot, u_char *packet, int seq, int *cc, struct timeval *tv)
{
	struct icmpv6_error *icp;
	u_char type, code;

	icp = get_icmp_hdr (packet);

	type = icp->head.type;
	code = icp->head.code;

	if ((type == ICMP6_TIMXCEED && code == ICMP6_TIMXCEED_INTRANS) ||
	    type == ICMP6_UNREACH) {
		struct ipv6hdr *hip;
		struct udphdr *up;

		
		hip = (struct ipv6hdr *) (icp + 1);
		/* dump_hex((char *)hip, 48); */
		if (hip->next_hdr == IPPROTO_UDP) {
			struct pkt_format *pkt;
			

			if (! inet_ntop(AF_INET6, &hip->saddr, printable_localaddr6, INET6_ADDRSTRLEN)) {
				perror ("inet_ntop of source address from sent packet: ");

				printable_localaddr6 [0] = 0;
			}

			up = (struct udphdr *)(hip+1);
			pkt = (struct pkt_format *) (up + 1);

			if (ntohl(pkt->ident) == ident &&
			    ntohl(pkt->seq) == seq) {
				*tv = pkt->tv;
				return (type == ICMP6_TIMXCEED? -1 : code+1);
			}
		}
	}

	if (verbose) {
		struct ipv6hdr *hip;
		unsigned long *lp;
		char printable_address[INET6_ADDRSTRLEN];
		int i;
		
		hip = (struct ipv6hdr *) (icp + 1);
		lp = (unsigned long *) (icp + 1);

		Printf("\n%d bytes from %s to %s", *cc,
		       inet_ntop(AF_INET6, &hip->saddr, printable_localaddr6, INET6_ADDRSTRLEN),
		       inet_ntop(AF_INET6, &hip->daddr, printable_address, INET6_ADDRSTRLEN));
		
		Printf(": icmp type %d (%s) code %d\n", type, pr_type(type), code);

		/* Dump IPv6 packet */
		for (i = sizeof(struct ipv6hdr); i < *cc ; i += sizeof(long))
			Printf("%2d: x%8.8lx\n", i, *lp++);
	}

	return(0);
}

void 
getnewhop6 (struct proto_ops *prot, int code, double difftime, int cc)
{
	struct sockaddr_in6	*from;
	struct in6_addr addr6, *lastaddr6;
	char printable_address[INET6_ADDRSTRLEN];
	char *hostnm;
	char localname [MAXHOSTNAMELEN];
	struct hostent *hp;
	
	
	from = (struct sockaddr_in6 *) prot->wherefrom;
	lastaddr6 = (struct in6_addr *) prot->lastaddr;
	
	if (memcmp (&from->sin6_addr, lastaddr6, sizeof (struct in6_addr))) {
		memcpy (prot->lastaddr, &from->sin6_addr, 
			                             sizeof (struct in6_addr));
		/* print info about the new hop found */
		if (nflag)
			Printf(" %s", inet_ntop(AF_INET6, &from->sin6_addr, 
					 printable_address, INET6_ADDRSTRLEN));
		else {
			if ((hp = gethostbyaddr((char *)&from->sin6_addr, 
				           sizeof(struct in6_addr), AF_INET6)))
				hostnm = hp->h_name;
			else
				hostnm = "* ";

			Printf(" %s (%s)", hostnm, inet_ntop(AF_INET6, &from->sin6_addr, printable_address, INET6_ADDRSTRLEN));
		}
		if (verbose) {
			if (!nflag){
				if (inet_pton(AF_INET6, printable_localaddr6, &addr6))
					 if ((hp = gethostbyaddr((char *)&addr6, sizeof(struct in6_addr), AF_INET6)))
					 	strncpy (printable_localaddr6, hp->h_name, INET6_ADDRSTRLEN);
			}
			printf (" %d bytes to %s", cc, printable_localaddr6);
		}
	}
	Printf("  %g ms", difftime);
	switch(code - 1) {
		case ICMP6_UNREACH_PORT:
			++got_there;
			break;
		case ICMP6_UNREACH_NOROUTE:
			++unreachable;
			Printf(" !N");
			break;
		case ICMP6_UNREACH_ADDRESS:
			++unreachable;
			Printf(" !H");
			break;
		case ICMP6_UNREACH_ADMIN:
			++unreachable;
			Printf(" !S");
			break;
	}
}


/*
 * Convert an ICMP "type" field to a printable string.
 */
char * 
pr_type(u_char t)
{
	static char *ttab1[] = {
		"Error",
		"Destination Unreachable",
		"Packet Too Big",
		"Time Exceeded in Transit",
		"Parameter Problem"
	};

	static char *ttab2[] = {
		"Echo Reply",
		"Echo Request",
		"Membership Query",
		"Membership Report",
		"Membership Reduction",
	};

	if (t < 5){
		return (ttab1[t]);
	}

	if (t >= 128 && t <= 132){
		return (ttab2[t-128]);
	}

	return("OUT-OF-RANGE");
}



struct proto_ops afinet6_prot = {
	AF_INET6,
	0,
	0,
	sizeof(struct sockaddr_in6),
	sizeof(struct pkt_format),
	NULL,
	NULL,
	NULL,
	init6,
	send_probe6,
	recv_packet6,
	getnewhop6,
	NULL};
	
