/*
 * DRET-IPv6	An implementation of the traceroute for the LINUX
 *		operating system.  
 *
 * Contacts:	
 *              INRIA      <Christophe.Diot@sophia.inria.fr>
 *              MASI       <Eric.Horlait@masi.ibp.fr>
 *              This software has been developped with the financial support
 *              of DRET (French Military Research Agency).
 *
 * Version:    $Id: traceroute.h,v 1.2 1996/12/11 14:09:40 eh Exp $ 
 *	
 * Authors:	
 *	        ANELLI Pascal	<anelli@masi.ibp.fr>
 *
 * Fixes:
 *
 *
 * Description:
 *
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *              This notice must be copied with the distributed package.
 *
 *		Portions of the software are derived from  various 
 *		networking code publicly available, mainly:
 *			NRL IPv6 code 
 *			NetBSD code from INRIA (Francis.Dupont@inria.fr)
 *			Pedro Roque's Linux version (Roque@di.fc.up.pt)
 *
 */				


/* 
 * That structure defines how traceroute must act with a given protocol, 
 * currently PF_INET and PF_INET6.
 */
   
struct proto_ops {
	int prot_af;			/* adress family */
	int s;				/* receive (icmp) socket file descriptor  */
	int sndsock;			/* send (udp) socket file descriptor */
	int sockaddrlen;		/* address length */
	int minlen_packet;		/* data lenght minimum send for a packet */
	struct sockaddr *whereto;	/* who to try to reach */
	struct sockaddr *wherefrom;	/* id of the source  or who to send a icmp packet */
	void *lastaddr;			/* last address found */
	int (*init)(char *target);	/* init protocol */
	void (*send_probe)(struct proto_ops *prot, u_char *outpacket, int seq, int ttl);
	int (*recv_packet)(struct proto_ops *prot, u_char *inpacket, int seq, int *cc, struct timeval *tv);
	void (*getnewhop)(struct proto_ops *prot, int code, double difftime, int cc);
	void (*options)(struct proto_ops *prot, u_char * outpacket);
#ifdef 0
	void (*pr_retip)(void *);		/* dump info on IP packet */
#endif
};
 
 /* 
 * Various constantes
 */
#define	MAXPACKET	(65536 - 60 - 8)/* max packet size */
#define	DEFDATALEN	(64 - 8)     	/* default data length */
#define	MAXIPLEN	60	/* Max header length 20 + 40 (options) */
#define	MAXICMPLEN	76	/* Max header length */
#define	MINICMPLEN	8	/* Min header length */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64
#endif

#ifndef FD_SET
#define NFDBITS         (8*sizeof(fd_set))
#define FD_SETSIZE      NFDBITS
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define Fprintf (void)fprintf
#define Sprintf (void)sprintf
#define Printf (void)printf

/* 
 * Command line options 
 */
#define F_IP4COMP		0x0200

/*
 * Some shortcuts 
 */


/* functions */
extern double	deltaT (struct timeval *, struct timeval *);
extern void	   tvsub (struct timeval *, struct timeval *);

/* external variable */
extern struct proto_ops afinet4_prot;
extern struct proto_ops afinet6_prot;

extern u_char inpacket[];		/* last inbound (icmp) packet */
extern u_char *outpacket;		/* last outbound (UDP) packet */
extern char   hostname[];

extern int datalen;			/* How much data */
extern int options;			/* socket options */
extern int verbose;
extern int nflag;			/* print addresses numerically */
extern int on;
extern int tos;

extern char *source;
extern int nprobes;
extern int max_ttl;
extern u_short port;	    /* start udp dest port # for probe packets */
extern int waittime;		/* time to wait for response (in seconds) */

extern u_short ident;		/* process ident */

extern int got_there;
extern int unreachable;
